package main

import "fmt"

type Feature interface {
	Fly(name string) bool
	Swim(name string) bool
}

type Duck struct {
}

// Duck  的相关方法声明为值接受者方法
func (d Duck) Fly(name string) bool {
	fmt.Println(name, "can fly")
	return true
}

func (d Duck) Swim(name string) bool {
	fmt.Println(name, "can swim")
	return true
}

type Swan struct {
}

// Swam 的相关方法声明为指针接收者方法
func (s *Swan) Fly(name string) bool {
	fmt.Println(name, "can fly")
	return true
}

func (s *Swan) Swim(name string) bool {
	fmt.Println(name, "can swim")
	return true
}

type Goose struct {
}

// Fly 该方法声明为值接受者，Go编译器默认声明指针类型的 Fly 方法
func (g Goose) Fly(name string) bool {
	fmt.Println(name, "can fly")
	return true
}

// Swim 该方法声明为指针接受者
func (g *Goose) Swim(name string) bool {
	fmt.Println(name, "can swim")
	return true
}

func doFly(name string, f Feature) {
	f.Fly(name)
}

type JustFly interface {
	Fly(name string) bool
}

func main() {
	var f Feature
	fmt.Printf("%T %#v\n", f, f) // <nil> <nil>

	// 实现多态
	f = Duck{}
	fmt.Printf("%T %#v\n", f, f) // main.Duck main.Duck{}

	f.Fly("tanghaoya")
	f.Swim("tanghaoya")

	f = &Swan{}                  // 只实现指针接收者方法，只能使用指针
	fmt.Printf("%T %#v\n", f, f) // main.Swan main.Swan{}

	f.Fly("tiane")
	f.Swim("tiane")

	// 值接收者方法自动添加指针接收者方法，因此可以使用指针赋值
	f = &Duck{}
	f.Fly("zhouheiya")
	f.Swim("zhouheiya")

	// 接口需要实现全部方法，否则不能称为接口
	//f = Goose{} // Cannot use 'Goose{}' (type Goose) as the type Feature Type does not implement 'Feature' as the 'Swim' method has a pointer receiver
	f = &Goose{}
	f.Fly("dabaie")
	f.Swim("dabaie")

	doFly("zhouhaiya", f)

	// 接口赋值给接口
	var js JustFly = f
	// 需要满足对应接口声明的方法和结构体的属性
	js.Fly("chicken")
	//js.Swim("chicken") 没有该方法

	// 接口断言
	goose := js.(*Goose)                        // 转为指针类型
	fmt.Printf("%T %#v\n", goose, goose)        //*main.Goose &main.Goose{}
	goose, ok := js.(*Goose)                    // 转为指针类型
	fmt.Printf("%T %#v %v\n", goose, goose, ok) //*main.Goose &main.Goose{} true

	swan, ok := js.(*Swan)
	// 类型不匹配，转化失败
	fmt.Printf("%T %#v\n", swan, ok) // *main.Swan false

	f = Duck{}
	js = f
	duck := js.(Duck)                          // 转为值类型
	fmt.Printf("%T %#v\n", duck, duck)         // main.Duck main.Duck{}
	duck, ok = js.(Duck)                       // 带是否转化成功标识
	fmt.Printf("%T %#v %#v\n", duck, duck, ok) // main.Duck main.Duck{} true

	// 类型查询，js.(type) 只能使用在 switch 中
	//switch js.(type) {
	switch v := js.(type) {
	case Duck:
		fmt.Println("duck")
		fmt.Printf("%T\n", v)
	//case Swan: // Impossible type switch case: 'Swan' does not implement 'JustFly'
	//	fmt.Println("swan")
	case *Swan: // Impossible type switch case: 'Swan' does not implement 'JustFly'
		fmt.Println("*swan")
	case Goose:
		fmt.Println("goose")
	default:
		fmt.Println("unknown")
	}
}
